/* eslint-env node */

const autoprefixer = require('autoprefixer');
const merge = require('merge-stream');
const purgecss = require('@fullhuman/postcss-purgecss');
const rimraf = require('rimraf');
const named = require('vinyl-named');
const webpack = require('webpack');
const webpackStream = require('webpack-stream');

const gulp = require('gulp');
const plugins = require('gulp-load-plugins');

const browser = require('browser-sync').create();

// Load all Gulp plugins into one variable
const $ = plugins({
  rename: {
    'gulp-touch-fd': 'touch'
  }
});

/// Configuration -------------------------------------------------------------

const CONFIG = {
  // Proxy target of the BrowserSync'server
  SERVER_PROXY: 'http://127.0.0.1:8000',

  // Port on which the BrowserSync'server will listen
  SERVER_PORT: 8090,

  // Paths to other assets which will be copied
  ASSETS_FILES: [
    {
      src: [
        'src/**/*',
        '!src/{img,js,scss}',
        '!src/{img,js,scss}/**/*'
      ],
      dest: ''
    },
    {
      // ForkAwesome
      src: 'node_modules/fork-awesome/fonts/*',
      dest: 'fonts/fork-awesome'
    }
  ],

  // Paths to images which will be compressed and copied
  IMAGES_FILES: [
    'src/img/**/*'
  ],

  // Paths to JavaScript entries which will be bundled
  JS_ENTRIES: [
    'src/js/app.js'
  ],

  // Paths to Sass files which will be compiled
  SASS_ENTRIES: [
    'src/scss/app.scss'
  ],

  // Paths of HTML files which should be parsed by UnCSS
  HTML_ENTRIES: [
    'src/*.html'
  ],

  // Paths to Sass libraries, which can then be loaded with @import
  SASS_INCLUDE_PATHS: [
    'node_modules'
  ],

  // Path to the build output, which can safely be be cleaned
  BUILD_PATH: 'dist'
};

/// CSS -----------------------------------------------------------------------

// Compile Sass into CSS.
function cssTranspile() {
  return gulp.src(CONFIG.SASS_ENTRIES)
    .pipe($.sourcemaps.init())
    .pipe($.sass({
      includePaths: CONFIG.SASS_INCLUDE_PATHS
    }).on('error', $.sass.logError))
    .pipe($.postcss([
      autoprefixer(),
      purgecss({
        content: CONFIG.HTML_ENTRIES,
        whitelistPatternsChildren: [/^modal/]
      })
    ]))
    .pipe($.sourcemaps.write('.'))
    .pipe(gulp.dest(`${CONFIG.BUILD_PATH}/css`))
    .pipe($.touch())
    .pipe(browser.reload({ stream: true }));
}

// Lint Sass files.
function cssLint() {
  return gulp.src('src/scss/**/*.scss')
    .pipe($.stylelint({
      failAfterError: true,
      reporters: [
        { formatter: 'verbose', console: true }
      ]
    }));
}

// Compress CSS files.
function cssMinify() {
  return gulp.src([
    `${CONFIG.BUILD_PATH}/css/*.css`,
    `!${CONFIG.BUILD_PATH}/css/*.min.css`
  ])
    .pipe($.cleanCss())
    .pipe($.rename({ suffix: '.min' }))
    .pipe(gulp.dest(`${CONFIG.BUILD_PATH}/css`));
}

exports.css = gulp.series(cssTranspile, cssMinify);

/// JavaScript ----------------------------------------------------------------

const webpackConfig = {
  devtool: 'source-map',
  mode: 'development',
  module: {
    rules: [
      {
        test: /.js$/,
        use: {
          loader: 'babel-loader',
          options: {
            presets: ['@babel/preset-env'],
            compact: false
          }
        }
      }
    ]
  },
  stats: {
    chunks: false,
    entrypoints: false
  }
};

// Bundle JavaScript module.
function jsBundle() {
  return gulp.src(CONFIG.JS_ENTRIES)
    .pipe(named())
    .pipe(webpackStream(webpackConfig, webpack))
    .pipe(gulp.dest(`${CONFIG.BUILD_PATH}/js`));
}

// Lint JavaScript source files.
function jsLint() {
  return gulp.src('src/js/**/*.js')
    .pipe($.eslint())
    .pipe($.eslint.format())
    .pipe($.eslint.failAfterError());
}

// Compress JavaScript files.
function jsMinify() {
  return gulp.src([
    `${CONFIG.BUILD_PATH}/js/*.js`,
    `!${CONFIG.BUILD_PATH}/js/*.min.js`
  ])
    .pipe($.terser())
    .pipe($.rename({ suffix: '.min' }))
    .pipe(gulp.dest(`${CONFIG.BUILD_PATH}/js`));
}

exports.js = gulp.series(jsBundle, jsMinify);

/// Other assets --------------------------------------------------------------

// Compress and copy images.
function images() {
  return gulp.src(CONFIG.IMAGES_FILES)
    .pipe($.imagemin({ progressive: true }))
    .pipe(gulp.dest(`${CONFIG.BUILD_PATH}/img`));
}

// Copy other assets files.
function assets() {
  return merge(CONFIG.ASSETS_FILES.map(
    item => gulp.src(item.src)
      .pipe(gulp.dest(`${CONFIG.BUILD_PATH}/${item.dest}`))
  ));
}

/// HTML files ----------------------------------------------------------------

/// BrowserSync servers -------------------------------------------------------

// Start a server with BrowserSync with the website.
function server(done) {
  browser.init({
    server: CONFIG.BUILD_PATH,
    ghostMode: false,
    notify: false
  });
  done();
}

/// Utils ---------------------------------------------------------------------

function clean(done) {
  rimraf(CONFIG.BUILD_PATH, done);
}

// Reload the BrowserSync server.
function reload(done) {
  browser.reload();
  done();
}

// Watch for changes to static assets, Sass and JavaScript.
function watch() {
  gulp.watch(['src/scss/**/*.scss'].concat(...CONFIG.HTML_ENTRIES),
    gulp.series(exports.css));
  gulp.watch('src/js/**/*.js',
    gulp.series(exports.js, reload));

  gulp.watch(CONFIG.IMAGES_FILES,
    gulp.series(images, reload));

  gulp.watch([].concat(...CONFIG.ASSETS_FILES.map(a => a.src)),
    gulp.series(assets, reload));
}

/// General tasks -------------------------------------------------------------

// Build and compress CSS, JavaScript and other assets.
exports.build = gulp.series(
  clean,
  gulp.parallel(cssTranspile, jsBundle),
  gulp.parallel(cssMinify, jsMinify),
  gulp.parallel(images, assets)
);

// Lint Sass and JavaScript sources.
exports.lint = gulp.parallel(cssLint, jsLint);

// Run a development server and watch for file changes.
exports.default = gulp.series(exports.build, server, watch);
